/* Restartable Sequences internal API.  Linux implementation.
   Copyright (C) 2020 Free Software Foundation, Inc.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */

#ifndef RSEQ_INTERNAL_H
#define RSEQ_INTERNAL_H

#include <sysdep.h>
#include <errno.h>
#include <kernel-features.h>
#include <stdio.h>
#include <sys/rseq.h>
#include <unistd.h>

#ifdef RSEQ_SIG

extern _Bool __rseq_active;

static inline void
rseq_active_init (void)
{
  __rseq_active = __access ("/etc/glibc.rseq.enabled", F_OK) == 0;
}

static inline void
rseq_register_current_thread (void)
{
  int ret;

  if (!__rseq_active)
    {
      __rseq_abi.cpu_id = RSEQ_CPU_ID_REGISTRATION_FAILED;
      return;
    }

  if (__rseq_abi.cpu_id != RSEQ_CPU_ID_UNINITIALIZED)
    __libc_fatal ("glibc fatal error: "
                  "rseq already initialized for this thread\n");
  ret = INTERNAL_SYSCALL_CALL (rseq, &__rseq_abi, sizeof (struct rseq),
                               0, RSEQ_SIG);
  if (INTERNAL_SYSCALL_ERROR_P (ret))
    {
      const char *msg = NULL;

      switch (INTERNAL_SYSCALL_ERRNO (ret))
        {
        case ENOSYS:    /* rseq system call not implemented.  */
        case EPERM:     /* rseq system call filtered by seccomp.  */
        case EACCES:    /* rseq system call filtered by seccomp.  */
          __rseq_abi.cpu_id = RSEQ_CPU_ID_REGISTRATION_FAILED;
          break;
        case EBUSY:
          msg = "glibc fatal error: rseq already registered for this thread\n";
          break;
        case EFAULT:
          msg = "glibc fatal error: rseq parameter is an invalid address\n";
          break;
        case EINVAL:
          msg = "glibc fatal error: rseq parameters are invalid\n";
          break;
        default:
          msg = "glibc fatal error: unexpected rseq errno\n";
          break;
        }
      if (msg != NULL)
        __libc_fatal (msg);
    }
}
#else /* RSEQ_SIG */
static inline void
rseq_active_init (void)
{
}

static inline void
rseq_register_current_thread (void)
{
}
#endif /* RSEQ_SIG */

#endif /* rseq-internal.h */
